/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000-2002 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

/*
 * Simple program to rotate Apache logs without having to kill the server.
 *
 * Contributed by Ben Laurie <ben@algroup.co.uk>
 *
 * 12 Mar 1996
 *
 * Ported to APR by Mladen Turk <mturk@mappingsoft.com>
 *
 * 23 Sep 2001
 *
 * Jan 2003 - Added rotate by calendar by Masato Kataoka
 *            for use with Orenosv/Orenosp.
 *            Based on rotatelogs.c in Apache 2.0.44.
 * Dec 2003 - Added post-process program by Masato Kataoka
 *
 */


#include "apr.h"
#include "apr_lib.h"
#include "apr_strings.h"
#include "apr_errno.h"
#include "apr_file_io.h"
#include "apr_file_info.h"
#include "apr_general.h"
#include "apr_time.h"

#if APR_HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if APR_HAVE_STRING_H
#include <string.h>
#endif
#if APR_HAVE_STRINGS_H
#include <strings.h>
#endif

#define BUFSIZE         65536
#define ERRMSGSZ        82

#ifndef MAX_PATH
#define MAX_PATH        1024
#endif


/* to enable enhancements for Orenosv/Orenosp */
#define CHAR_SPECIAL 1

#define MYVERSION "2.0.44-oreno-0.3"

/*
 * How to build rotatelogs.exe on Windows
  http://httpd.apache.org/docs-2.0/platform/win_compiling.html  
  get awk.exe (see doc in the above)
  
  >vcvars32.bat
  >nmake /f Makefile.win {_apached|_apacher}

  >cd support
  >NMAKE -nologo -f rotatelogs.mak  CFG="rotatelogs - Win32 Debug" RECURSE=0 
  >NMAKE -nologo -f rotatelogs.mak  CFG="rotatelogs - Win32 Release" RECURSE=0 

  if you want to debug, edit rotatelogs.mak to enable debug info in Release.
 */




#ifdef _WIN32
/* MSVCRT's system() can't handle space-containing arguments */
int mysystem(const char *command);
#define system(cmd) mysystem((cmd))
#endif

/*
 * invoke post-process program
 * TODO : rewrite using APR functions...
 */
#include <stdlib.h>
static int
invoke_postprocess(char *logfname)
{
  char temp[512], tms[32];
  int rc;
  char *p1, *p2;
  int log_cmd = 0;
  int do_2G = 0;  /* postprocess more than 2GB log file also */

  if (p1 = getenv("ROTATELOGS_POSTPROCESS")) {
    FILE *log = NULL;
    time_t begin;

    if (p2 = getenv("ROTATELOGS_POSTPROC_OPTS")) {
      if (strstr(p2, "-log"))
	log_cmd = 1;
      if (strstr(p2, "-do2g"))
	do_2G = 1;
    }

    if (!do_2G) {
        int okay = 0;
	apr_finfo_t finfo;
	apr_pool_t *pool;
	
	apr_pool_create(&pool, NULL);
	if (apr_stat(&finfo, logfname, APR_FINFO_SIZE, pool) == APR_SUCCESS) {
	    if (finfo.size < ((apr_size_t)2) * 1024*1024*1024)
	      okay = 1;
	}
	apr_pool_destroy(pool);
	if (!okay)
	    return 0;
    }

    _flushall(); /* per MS CRT system() man page */
    /*
     * invoking :
     * "%ROTATELOGS_POSTPROCESS%" "<LOGFNAME>"
     */
    sprintf(temp, "\"%s\" \"%s\"", p1, logfname);
    if (log_cmd) {
      strcat(temp, " >> rotatelogs-postproc.out 2>&1");
    }

    time(&begin);
    
    rc = system(temp);

    if (log_cmd) {
      struct tm tm;
      time_t now;
      int elapse;
      
      time(&now);
      elapse = now - begin;
      memcpy(&tm, localtime(&begin), sizeof(tm));
      strftime(tms, sizeof(tms), "%y/%m/%d %H:%M:%S", &tm);

      log = fopen("rotatelogs-postproc.log", "at");
      if (log) {
        fprintf(log, "%s [%s] took %d secs, returned %d, errno:%d\n",
		tms, temp, elapse, rc, errno);
        fclose(log);
      }
    }
  }
  return 0;
}

static thread_start(void *arg)
{
  invoke_postprocess((char *)arg);
  return 0;
}

start_postprocess(char *logfname)
{
  unsigned long h, id;
  
  h = _beginthreadex(NULL, 0, (unsigned (__stdcall *)(void *))thread_start,
		     (void *)logfname, 0, &id);
  if (h == 0) {
    return -1;
  }
  
  CloseHandle((HANDLE)h);
  return 0;
}


/*
 * find begining of period in apr_time_t (i.e., in UTC)
 *
 * for APR time functions see
 *   http://apr.apache.org/docs/apr/group__APR__Time.html
 */
static int
find_bo_period(int caldef, apr_time_t now, apr_time_t *result)
{
    apr_time_exp_t tm_now, tm_begin;
    apr_time_t begin;
    
    /* convert time_t to struct tm(local) */
    apr_time_exp_lt(&tm_now, now);

    if (caldef == 1) { /* month */
	tm_begin = tm_now;
	tm_begin.tm_usec = 0;
	tm_begin.tm_sec = 0;
	tm_begin.tm_min = 0;
	tm_begin.tm_hour = 0;
	tm_begin.tm_mday = 1;

	/* convert struct tm(local) to time_t(UTC) */
	/* tm_mday is an essential field.  we don't need to recalc
	 * other non-essential fields.
	 */
	apr_time_exp_gmt_get(&begin, &tm_begin);

    } else if (caldef == 2 || caldef == 3) { /* week, weekmon */
	int diff_sec;
	int days;
	/* tm_wday is Sun=0, Mon=1, ... */
	if (caldef == 2)
	    days = tm_now.tm_wday;
	else /* start on Monday */
	    days = ((tm_now.tm_wday == 0) ? 6 : (tm_now.tm_wday - 1));
	
	diff_sec = days * 60 * 60 * 24;
	diff_sec += tm_now.tm_hour * 60 * 60 + tm_now.tm_min * 60 +
	            tm_now.tm_sec;
	
	begin = now - apr_time_from_sec(diff_sec);
    } else if (caldef == 4) { /* day */
	int diff_sec = 0;
	diff_sec += tm_now.tm_hour * 60 * 60 + tm_now.tm_min * 60 +
	            tm_now.tm_sec;
	begin = now - apr_time_from_sec(diff_sec);
    }
    *result = begin;
    return 0;
}


int main (int argc, const char * const argv[])
{
    char buf[BUFSIZE], buf2[MAX_PATH], errbuf[ERRMSGSZ];
    int tLogEnd = 0, tRotation = 0, utc_offset = 0;
    /* --- rotate by size --- */
    /* unsigned int sRotation = 0; */
    apr_size_t sRotation = 0;
#ifdef CHAR_SPECIAL
    /* --- rotate by calendar -- */
    int cRotation = 0; /* rotate by calendar */
    apr_time_t currbegin; /* begining of current period */
    /* --- end --- */
#endif
    int nMessCount = 0;
    apr_size_t nRead, nWrite;
    int use_strftime = 0;
    int now = 0;
    const char *szLogRoot;
    apr_file_t *f_stdin, *nLogFD = NULL, *nLogFDprev = NULL;
    apr_pool_t *pool;
    char *ptr = NULL;
    char prevLogFName[MAX_PATH];

    apr_app_initialize(&argc, &argv, NULL);
    atexit(apr_terminate);

#ifdef CHAR_SPECIAL
    if (argc >= 2 && !strcmp(argv[1], "-test")) { /* test */
	apr_time_t begin;
	apr_time_exp_t e;
	apr_size_t rs;
	char temp[64];
	
	find_bo_period(1, apr_time_now(), &begin);
	apr_ctime(temp, begin);
	printf("by month [%s]\n", temp);
	find_bo_period(2, apr_time_now(), &begin);
	apr_ctime(temp, begin);
	printf("by week [%s]\n", temp);
	
	find_bo_period(3, apr_time_now(), &begin);
	apr_ctime(temp, begin);
	printf("by weekmon [%s]\n", temp);
	
	apr_time_exp_lt(&e, begin);
	apr_strftime(temp, &rs, sizeof(temp), "%Y%m%d-%H%M%S.log", &e);
	printf("by weekmon [%s]\n", temp);
	
	find_bo_period(4, apr_time_now(), &begin);
	apr_ctime(temp, begin);
	printf("by day [%s]\n", temp);
#if 0
	start_postprocess("c:/log files/access-20031209-115800.log");
#endif
	exit(0);
    }
#endif

    apr_pool_create(&pool, NULL);
    if (argc < 3 || argc > 4) {
#ifdef CHAR_SPECIAL
        fprintf(stderr,
		"rotatelogs-char " MYVERSION "\n"
		"Usage: %s <logfile-specifier> <mode-specific-options>\n"
                "Mode-1 <rotation time in seconds> [offset minutes from UTC]\n"
		"Mode-2 <rotation size in Megabytes or Gigabytes>\n"
		"Mode-3 {-cal:month | -cal:week | -cal:weekmon | -cal:day}\n"
		"\n", argv[0]);
#else
        fprintf(stderr,
                "Usage: %s <logfile> <rotation time in seconds> "
                "[offset minutes from UTC] or <rotation size in megabytes>\n\n",
                argv[0]);
#endif
#ifdef OS2
        fprintf(stderr,
                "Add this:\n\nTransferLog \"|%s.exe /some/where 86400\"\n\n",
                argv[0]);
#else
        fprintf(stderr,
                "Add this:\n\nTransferLog \"|%s /some/where 86400\"\n\n",
                argv[0]);
        fprintf(stderr,
                "or \n\nTransferLog \"|%s /some/where 5M\"\n\n", argv[0]);
#endif
        fprintf(stderr,
                "to httpd.conf. The generated name will be /some/where.nnnn "
                "where nnnn is the\nsystem time at which the log nominally "
                "starts (N.B. if using a rotation time,\nthe time will always "
                "be a multiple of the rotation time, so you can synchronize\n"
                "cron scripts with it). At the end of each rotation time or "
                "when the file size\nis reached a new log is started.\n");
        exit(1);
    }

    szLogRoot = argv[1];

    ptr = strchr (argv[2], 'M');
    if (ptr) {
        if (*(ptr+1) == '\0') {
            sRotation = atoi(argv[2]) * 1048576;
        }
        if (sRotation == 0) {
            fprintf(stderr, "Invalid rotation size parameter\n");
            exit(1);
        }
    } 
#if 0
    else if (ptr = strchr (argv[2], 'G')) {
        if (*(ptr+1) == '\0') {
            sRotation = atoi(argv[2]) * 1024*1024*1024;
        }
        if (sRotation == 0) {
            fprintf(stderr, "Invalid rotation size parameter\n");
            exit(1);
        }
    }
#endif
#ifdef CHAR_SPECIAL
    else if (!memcmp(argv[2], "-cal:", 5)) {
	const char *p1 = &argv[2][5];
	/*
	 * 1: month
	 * 2: week (begins on Sunday)
	 * 3: weekmon (begins on Monday)
	 * 4: day
	 */
	if (!strcmp(p1, "month")) {
	    cRotation = 1;
	} else if (!strcmp(p1, "week")) {
	    cRotation = 2;
	} else if (!strcmp(p1, "weekmon")) {
	    cRotation = 3;
	} else if (!strcmp(p1, "day")) {
	    cRotation = 4;
	} else {
	    fprintf(stderr, "Invalid calendar specififer\n");
	    exit(1);
	}
    }
#endif
    else {
        if (argc >= 4) {
            utc_offset = atoi(argv[3]) * 60;
        }
        tRotation = atoi(argv[2]);
        if (tRotation <= 0) {
            fprintf(stderr, "Rotation time must be > 0\n");
            exit(6);
        }
    }

    use_strftime = (strchr(szLogRoot, '%') != NULL);
    if (apr_file_open_stdin(&f_stdin, pool) != APR_SUCCESS) {
        fprintf(stderr, "Unable to open stdin\n");
        exit(1);
    }

    for (;;) {
        nRead = sizeof(buf);

        if (apr_file_read(f_stdin, buf, &nRead) != APR_SUCCESS)
            exit(3);

        if (tRotation) {
            now = (int)(apr_time_now() / APR_USEC_PER_SEC) + utc_offset;
            if (nLogFD != NULL && now >= tLogEnd) {
                nLogFDprev = nLogFD;
                nLogFD = NULL;
		strcpy(prevLogFName, buf2);
            }
        }
        else if (sRotation) {
            apr_finfo_t finfo;
            apr_off_t current_size = -1;

            if ((nLogFD != NULL) && 
                (apr_file_info_get(&finfo, APR_FINFO_SIZE, nLogFD) == APR_SUCCESS)) {
                current_size = finfo.size;
            }

            if (current_size > sRotation) {
                nLogFDprev = nLogFD;
                nLogFD = NULL;
		strcpy(prevLogFName, buf2);
            }
        }
#ifdef CHAR_SPECIAL
	else if (cRotation) {
	    apr_time_t begin;
	    
	    find_bo_period(cRotation, apr_time_now(), &begin);

	    /* note about time comparison between currbegin and begin.
	     * find_bo_period is not very accurate so we should round it
	     * up to a few seconds...
	     */
	    if (nLogFD != NULL &&
		apr_time_sec(begin) > apr_time_sec(currbegin) + 1) {
		
                nLogFDprev = nLogFD;
                nLogFD = NULL;
		strcpy(prevLogFName, buf2);
	    }
	}
#endif
        else {
            fprintf(stderr, "No rotation time or size specified\n");
            exit(2);
        }

        if (nLogFD == NULL) {
	    /*
	     * open initial new log file, or
	     * close current log file and open a new log file
	     */
            int tLogStart;

	    /*
	     * record log start time in tLogStart
	     */
            if (tRotation)
                tLogStart = (now / tRotation) * tRotation;
#ifdef CHAR_SPECIAL
	    else if (cRotation) {
		apr_time_t begin;
	    
		find_bo_period(cRotation, apr_time_now(), &begin);
		currbegin = begin;

		/* in rotate by time mode (tRotation), tLogStart is
		 * is a value already offsetted by utc_offset.
		 * in cRotation mode, it's not.  it's in UTC.
		 */
		tLogStart = (int)apr_time_sec(currbegin);
	    }
#endif
            else
                tLogStart = (int)apr_time_sec(apr_time_now());

	    /*
	     * generate a file name for new log file in "buf2"
	     */
            if (use_strftime) {
                apr_time_t tNow = apr_time_from_sec(tLogStart);
                apr_time_exp_t e;
                apr_size_t rs;

#ifdef CHAR_SPECIAL
		if (cRotation)
		    apr_time_exp_lt(&e, tNow);
		else
		    apr_time_exp_gmt(&e, tNow);
#else
		apr_time_exp_gmt(&e, tNow);
#endif
                apr_strftime(buf2, &rs, sizeof(buf2), szLogRoot, &e);
            }
            else {
                sprintf(buf2, "%s.%010d", szLogRoot, tLogStart);
            }
            tLogEnd = tLogStart + tRotation;

            apr_file_open(&nLogFD, buf2, APR_READ | APR_WRITE | APR_CREATE | APR_APPEND,
                          APR_OS_DEFAULT, pool);
            if (nLogFD == NULL) {
                /* Uh-oh. Failed to open the new log file. Try to clear
                 * the previous log file, note the lost log entries,
                 * and keep on truckin'. */
                if (nLogFDprev == NULL) {
                    fprintf(stderr, "1 Previous file handle doesn't exists %s\n", buf2);
                    exit(2);
                }
                else {
                    nLogFD = nLogFDprev;
                    sprintf(errbuf,
                            "Resetting log file due to error opening "
                            "new log file. %10d messages lost.\n",
                            nMessCount);
                    nWrite = strlen(errbuf);
                    apr_file_trunc(nLogFD, 0);
                    if (apr_file_write(nLogFD, errbuf, &nWrite) != APR_SUCCESS) {
                        fprintf(stderr, "Error writing to the file %s\n", buf2);
                        exit(2);
                    }
                }
            }
            else if (nLogFDprev) {
		/*
		 * open succeeded.  now close old log file
		 */
                apr_file_close(nLogFDprev);
		start_postprocess(prevLogFName);
            }
            nMessCount = 0;
        }
        nWrite = nRead;
        apr_file_write(nLogFD, buf, &nWrite);
        if (nWrite != nRead) {
            nMessCount++;
            sprintf(errbuf,
                    "Error writing to log file. "
                    "%10d messages lost.\n",
                    nMessCount);
            nWrite = strlen(errbuf);
            apr_file_trunc(nLogFD, 0);
            if (apr_file_write(nLogFD, errbuf, &nWrite) != APR_SUCCESS) {
                fprintf(stderr, "Error writing to the file %s\n", buf2);
                exit(2);
            }
        }
        else {
            nMessCount++;
        }
    }
    /* Of course we never, but prevent compiler warnings */
    return 0;
}


#ifdef _WIN32
int mysystem(const char *command)
{
  BOOL rc;
  int ec;
  int crflags = 0;
  STARTUPINFO si;
  PROCESS_INFORMATION pi;
  char cmdline[1024];
  int exitcode;
  
  memset(&pi, 0, sizeof(pi));
  memset(&si, 0, sizeof(si));
  si.cb = sizeof(si);

  /*
   * Note: we need to close the whole command line (which may already contain
   * double quotes) with double quotes for cmd.exe to work.
   * cmd.c /c ""c:/progs/zip.exe" "c:/log files/log.zip" "c:/log files/log""
   */
  sprintf(cmdline, "cmd.exe /c \"%s\"", command);
  
  rc = CreateProcess(
		     NULL,
		     cmdline,
		     NULL, NULL, /* sec attr for proc and thread */
		     FALSE, /* inherit inheritable handle */
		     crflags,
		     NULL, /* env */
		     NULL, /* current working dir */
		     &si,
		     &pi);
  if (!rc) {
    ec = GetLastError();
    return -1;
  }
  CloseHandle(pi.hThread);


  /*
   * now wait for the process to finish
   */
  rc = WaitForSingleObject(pi.hProcess, INFINITE);
  if (rc == WAIT_OBJECT_0) {
    GetExitCodeProcess(pi.hProcess, &exitcode);
    /* exitcode == STILL_ACTIVE --> BAD! */
    CloseHandle(pi.hProcess);
  } else {
    /* WAIT_TIMEOUT or WAIT_ABANDONED ... */
    CloseHandle(pi.hProcess);
  }
  /* XXX wait(2) */
  return 0;
}

#endif
